from lxml.html import fromstring
from plone.app.blocks import utils
from plone.app.blocks.layoutbehavior import ILayoutAware
from plone.tiles.data import ANNOTATIONS_KEY_PREFIX
from zope.annotation.interfaces import IAnnotations


def onLayoutEdited(obj, event):
    """
    need to get the layout because you need to know what are
    acceptible storage values
    """
    lookup = ILayoutAware(obj)
    layout = lookup.content_layout()

    if not layout:
        return

    tree = fromstring(layout)
    tile_keys = []
    for el in utils.bodyTileXPath(tree):
        tile_url = el.attrib.get("data-tile", "")
        if "plone.app.standardtiles.field" in tile_url:
            continue
        tile_keys.append(
            ANNOTATIONS_KEY_PREFIX + "." + tile_url.split("?")[0].split("/")[-1]
        )

    annotations = IAnnotations(obj)
    for key in list(annotations.keys()):
        if key.startswith(ANNOTATIONS_KEY_PREFIX) and key not in tile_keys:
            del annotations[key]



from plone.app.blocks.testing import BLOCKS_FUNCTIONAL_TESTING
from plone.app.blocks.utils import resolve
from plone.app.blocks.utils import resolveResource

import unittest


class TestUtils(unittest.TestCase):
    def test_resolve_utf8_unicode(self):
        content_layout = """
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html data-layout="./@@default-site-layout">
  <body>
    <h1>Ä</h1>
  </body>
</html>"""
        text = resolve("", content_layout).xpath("//h1")[0].text
        self.assertEqual("Ä", text)

    def test_resolve_utf8_bytestring(self):
        """Test fix for issue where layouts with non-ascii characters were
        not properly parsed resulting in double encoding
        """
        content_layout = """
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html data-layout="./@@default-site-layout">
  <body>
    <h1>Ä</h1>
  </body>
</html>""".encode()
        text = resolve("", content_layout).xpath("//h1")[0].text
        self.assertEqual("Ä", text)


class TestUtilsFunctional(unittest.TestCase):
    layer = BLOCKS_FUNCTIONAL_TESTING

    def test_resolve_resource(self):
        layout = resolveResource("/++sitelayout++testlayout1/site.html")
        self.assertIn("Layout panel 1", layout)

    def test_resolve_resource_with_query(self):
        layout = resolveResource("/++sitelayout++testlayout1/site.html?ajax_load=1")
        self.assertIn("Layout panel 1", layout)

    def test_resolve_resource_with_html_content(self):
        # make sure we can handle content containing other ++ urls
        layout = resolveResource(
            "/?content=%3Cp%3ELorem%20ipsum%20dolor%20sit%20amet,%20consectetuer%20adipiscing%20elit.%20Sed%20posuere%20interdum%20sem.%20Quisque%20ligula%20eros%20ullamcorper%20quis,%20lacinia%20quis%20facilisis%20sed%20sapien.%3C/p%3E%0A%3Cscript%20type=%22application/javascript%22%20id=%22protect-script%22%20src=%22../../../++resource++protect.js"
        )
        self.assertIn("Plone Open Source CMS", layout)



from plone.app.blocks.testing import BLOCKS_FIXTURE
from plone.app.linkintegrity.handlers import modifiedContent
from plone.app.linkintegrity.interfaces import IRetriever
from plone.app.linkintegrity.parser import extractLinks
from plone.app.testing import IntegrationTesting
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.base.interfaces import IEditingSchema
from plone.dexterity.fti import DexterityFTI
from plone.registry.interfaces import IRegistry
from plone.tiles import Tile
from plone.uuid.interfaces import IUUID
from zope.component import adapter
from zope.component import getUtility
from zope.configuration import xmlconfig
from zope.interface import implementer
from zope.interface import Interface

import pkg_resources
import unittest


try:
    pkg_resources.get_distribution("plone.app.contenttypes")
except pkg_resources.DistributionNotFound:
    HAS_PLONE_APP_CONTENTTYPES = False
else:
    HAS_PLONE_APP_CONTENTTYPES = True


class ITestTile(Interface):
    """test marker"""


class TestTile(Tile):
    def __call__(self):
        return (
            f"<a href=\"resolveuid/{self.request.form.get('uid')}\">internal link</a>"
        )


@implementer(IRetriever)
@adapter(TestTile)
class TestTileRetriever:
    def __init__(self, context):
        self.context = context

    def retrieveLinks(self):
        content = self.context()
        return set(extractLinks(content))


class TestTilesLayer(PloneSandboxLayer):
    defaultBases = (BLOCKS_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        xmlconfig.string(
            """\
<configure
    xmlns="http://namespaces.zope.org/zope"
    xmlns:plone="http://namespaces.plone.org/plone"
    i18n_domain="plone.app.blocks">

  <include package="plone.tiles" file="meta.zcml" />

  <plone:tile
      name="test.tile"
      title="Test Tile"
      add_permission="cmf.ModifyPortalContent"
      class="plone.app.blocks.tests.test_linkintegrity.TestTile"
      permission="zope2.View"
      for="*"
      />

  <adapter
      factory="plone.app.blocks.tests.test_linkintegrity.TestTileRetriever"
      />

</configure>
""",
            context=configurationContext,
        )


BLOCKS_TILES_LINKINTEGRITY_FIXTURE = TestTilesLayer()
BLOCKS_TILES_LINKINTEGRITY_INTEGRATION_TESTING = IntegrationTesting(
    bases=(BLOCKS_TILES_LINKINTEGRITY_FIXTURE,),
    name="Blocks:Tiles:LinkIntegrity:Integration",
)


class TestLinkIntegrity(unittest.TestCase):
    layer = BLOCKS_TILES_LINKINTEGRITY_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self.registry = getUtility(IRegistry)
        self.maxDiff = None

        editing_settings = self.registry.forInterface(IEditingSchema, prefix="plone")
        editing_settings.enable_link_integrity_checks = True

        fti = DexterityFTI(
            "MyDocument",
            global_allow=True,
            behaviors=(
                "plone.app.dexterity.behaviors.metadata.IBasic",
                "plone.app.blocks.layoutbehavior.ILayoutAware",
            ),
        )
        self.portal.portal_types._setObject("MyDocument", fti)

        setRoles(self.portal, TEST_USER_ID, ("Manager",))
        self.portal.invokeFactory("Folder", "f1", title="Folder 1")
        self.folder = self.portal["f1"]
        self.folder.invokeFactory("MyDocument", "d1", title="Document 1")
        self.doc1 = self.folder["d1"]
        self.folder.invokeFactory("MyDocument", "d2", title="Document 2")
        self.doc2 = self.folder["d2"]

        # set customContentLayout to @@test_tile with internal Link
        self.doc1.customContentLayout = f'<html><body><div data-tile="./@@test.tile?uid={IUUID(self.doc2)}"/></body></html>'

    def test_linkintegrity(self):
        # TODO TEST
        pass

    def test_copy_paste(self):
        # see
        _cp = self.folder.manage_copyObjects(
            [
                "d1",
            ]
        )
        self.folder.manage_pasteObjects(_cp)

        self.assertTrue("copy_of_d1" in self.folder)

        _cp = self.portal.manage_copyObjects(
            [
                "f1",
            ]
        )
        self.portal.manage_pasteObjects(_cp)

        self.assertTrue("copy_of_f1" in self.portal)
        self.assertTrue("d1" in self.portal["copy_of_f1"])



from lxml import etree
from lxml import html
from plone.app.blocks import panel
from plone.app.blocks import tiles
from plone.base.utils import safe_bytes
from plone.tiles import esi
from plone.tiles.interfaces import ESI_HEADER
from plone.transformchain.interfaces import ITransform
from repoze.xmliter.serializer import XMLSerializer
from repoze.xmliter.utils import getHTMLSerializer
from zope.interface import implementer

import logging
import re


logger = logging.getLogger(__name__)


@implementer(ITransform)
class DisableParsing:
    """A no-op transform which sets flags to stop plone.app.blocks
    transformations. You may register this for a particular published
    object or request as required. By default, it's registered for ESI-
    rendered tiles when they are fetched via ESI.
    """

    order = 8000

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def transformBytes(self, result, encoding):
        self.request.set("plone.app.blocks.disabled", True)

    def transformUnicode(self, result, encoding):
        self.request.set("plone.app.blocks.disabled", True)

    def transformIterable(self, result, encoding):
        self.request.set("plone.app.blocks.disabled", True)


@implementer(ITransform)
class ParseXML:
    """First stage in the 8000's chain: parse the content to an lxml tree
    encapsulated in an XMLSerializer.

    The subsequent steps in this package will assume their result inputs are
    XMLSerializer iterables, and do nothing if it is not. This also gives us
    the option to parse the content here, and if we decide it's not HTML,
    we can avoid trying to parse it again.
    """

    order = 8000

    # Tests set this to True
    pretty_print = False

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def transformBytes(self, result, encoding):
        return self.transformIterable([result], encoding)

    def transformUnicode(self, result, encoding):
        return self.transformIterable([result], encoding)

    def transformIterable(self, result, encoding):
        if self.request.get("plone.app.blocks.disabled", False):
            return

        content_type = self.request.response.getHeader("Content-Type")
        if content_type is None or not content_type.startswith("text/html"):
            return

        contentEncoding = self.request.response.getHeader("Content-Encoding")
        if contentEncoding and contentEncoding in (
            "zip",
            "deflate",
            "compress",
        ):
            return

        try:
            # Fix layouts with CR[+LF] line endings not to lose their heads
            # (this has been seen with downloaded themes with CR[+LF] endings)
            # The html serializer much prefers only bytes, no unicode/text,
            # and it return a serializer that returns bytes.
            # So we start with ensuring all items in the iterable are bytes.
            iterable = [
                re.sub(b"&#13;", b"\n", re.sub(b"&#13;\n", b"\n", safe_bytes(item)))
                for item in result
                if item
            ]
            result = getHTMLSerializer(
                iterable, pretty_print=self.pretty_print, encoding=encoding
            )
            # Fix XHTML layouts with where etree.tostring breaks <![CDATA[
            if any([b"<![CDATA[" in item for item in iterable]):
                result.serializer = html.tostring

            self.request["plone.app.blocks.enabled"] = True
            return result
        except (AttributeError, TypeError, etree.ParseError) as e:
            logger.error(e)


@implementer(ITransform)
class MergePanels:
    """Find the site layout and merge panels."""

    order = 8100

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def transformBytes(self, result, encoding):
        return

    def transformUnicode(self, result, encoding):
        return

    def transformIterable(self, result, encoding):
        if not self.request.get("plone.app.blocks.enabled", False) or not isinstance(
            result, XMLSerializer
        ):
            return

        tree = panel.merge(self.request, result.tree)
        if tree is None:
            return

        # Set a marker in the request to let subsequent steps know the merging
        # has happened
        self.request["plone.app.blocks.merged"] = True

        result.tree = tree

        # Fix serializer when layout has changed doctype from XHTML to HTML
        if result.tree.docinfo.doctype and "XHTML" not in result.tree.docinfo.doctype:
            result.serializer = html.tostring

        return result


@implementer(ITransform)
class IncludeTiles:
    """Turn a panel-merged page into the final composition by including tiles.
    Assumes the input result is an lxml tree and returns an lxml tree for
    later serialization.
    """

    order = 8500

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def transformBytes(self, result, encoding):
        return

    def transformUnicode(self, result, encoding):
        return

    def transformIterable(self, result, encoding):
        if not self.request.get("plone.app.blocks.enabled", False) or not isinstance(
            result, XMLSerializer
        ):
            return

        result.tree = tiles.renderTiles(self.request, result.tree)
        return result


@implementer(ITransform)
class ESIRender:
    """If ESI rendering was used, render the page down to a format that allows
    ESI to work.
    """

    order = 9900

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def transformBytes(self, result, encoding):
        if self.request.getHeader(ESI_HEADER, "false").lower() != "true":
            return

        return esi.substituteESILinks([result])

    def transformUnicode(self, result, encoding):
        if self.request.getHeader(ESI_HEADER, "false").lower() != "true":
            return

        return esi.substituteESILinks([result.encode(encoding, "ignore")])

    def transformIterable(self, result, encoding):
        if self.request.getHeader(ESI_HEADER, "false").lower() != "true":
            return
        result = "".join(str(result))
        transformed = esi.substituteESILinks(result)
        if transformed != result:
            self.request.response.setHeader("X-Esi", "1")
        return transformed



from plone.app.blocks.layoutbehavior import ILayoutAware
from plone.app.blocks.layoutbehavior import ILayoutBehaviorAdaptable
from plone.app.blocks.layoutbehavior import LayoutAwareTileDataStorage
from plone.app.blocks.testing import BLOCKS_FUNCTIONAL_TESTING
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.textfield import RichText
from plone.app.textfield import RichTextValue
from plone.dexterity.fti import DexterityFTI
from plone.registry.interfaces import IRegistry
from plone.rfc822.interfaces import IPrimaryField
from plone.supermodel.model import Schema
from plone.tiles.interfaces import ITileType
from plone.tiles.type import TileType
from plone.uuid.interfaces import IUUID
from zope.component import getUtility
from zope.component import provideUtility
from zope.interface import alsoProvides

import pkg_resources
import unittest


try:
    pkg_resources.get_distribution("plone.app.contenttypes")
except pkg_resources.DistributionNotFound:
    HAS_PLONE_APP_CONTENTTYPES = False
else:
    HAS_PLONE_APP_CONTENTTYPES = True


class TestLayoutBehavior(unittest.TestCase):
    layer = BLOCKS_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self.registry = getUtility(IRegistry)
        self.maxDiff = None

        fti = DexterityFTI(
            "MyDocument",
            global_allow=True,
            behaviors=(
                "plone.app.dexterity.behaviors.metadata.IBasic",
                "plone.app.blocks.layoutbehavior.ILayoutAware",
            ),
        )
        self.portal.portal_types._setObject("MyDocument", fti)

        setRoles(self.portal, TEST_USER_ID, ("Manager",))
        self.portal.invokeFactory("Folder", "f1", title="Folder 1")
        self.portal["f1"].invokeFactory("MyDocument", "d1", title="Document 1")
        setRoles(self.portal, TEST_USER_ID, ("Member",))

        self.behavior = ILayoutAware(self.portal["f1"]["d1"])
        self.assertTrue(ILayoutBehaviorAdaptable.providedBy(self.behavior.context))

    def test_custom_content_layout(self):
        from plone.app.blocks.layoutviews import ContentLayoutView

        self.behavior.customContentLayout = (
            '<html><body><a href="{:s}"></a></body></html>'.format(
                "resolveuid/{:s}".format(IUUID(self.portal["f1"]))
            )
        )
        rendered = ContentLayoutView(self.portal["f1"]["d1"], self.request)()

        # Test that UUID is resolved by outputfilters
        self.assertNotIn(IUUID(self.portal["f1"]), rendered)
        self.assertIn(self.portal["f1"].absolute_url(), rendered)

    def test_content(self):
        tile = "plone.app.tiles.demo.transient/demo"
        self.behavior.content = """\
<html>
<body>
<div data-tile="@@plone.app.tiles.demo.transient/demo"
data-tiledata='{"message": "Hello World!"}' />
</body>
</html>
"""
        view = self.portal["f1"]["d1"].restrictedTraverse(tile)()
        self.assertEqual(
            view, "<html><body><b>Transient tile Hello World!</b></body></html>"
        )

        storage = LayoutAwareTileDataStorage(
            self.portal["f1"]["d1"], self.layer["request"]
        )
        self.assertEqual(len(storage), 1)
        self.assertEqual(list(storage), [tile])
        self.assertEqual(storage[tile]["message"], "Hello World!")

        data = storage[tile]
        data["message"] = "Foo bar!"
        storage[tile] = data

        view = self.portal["f1"]["d1"].restrictedTraverse(tile)()
        self.assertEqual(
            view, "<html><body><b>Transient tile Foo bar!</b></body></html>"
        )

    def test_content_richtext(self):
        class IRichTextTile(Schema):
            html = RichText()

        alsoProvides(IRichTextTile["html"], IPrimaryField)

        provideUtility(
            TileType(
                name="plone.app.blocks.richtext",
                title="plone.app.blocks.richtext",
                add_permission="cmf.ModifyPortalContent",
                view_permission="zope2.View",
                schema=IRichTextTile,
            ),
            provides=ITileType,
            name="plone.app.blocks.richtext",
        )

        self.behavior.content = """\
<html>
<body>
<div data-tile="@@plone.app.blocks.richtext/demo"
data-tiledata='{"content-type": "text/html"}'>
<div><p>Hello World!</p></div>
</div>
</body>
</html>
"""
        storage = LayoutAwareTileDataStorage(
            self.portal["f1"]["d1"], self.layer["request"]
        )
        data = storage["@@plone.app.blocks.richtext/demo"]

        self.assertIn("html", data)
        self.assertIsInstance(data["html"], RichTextValue)
        self.assertEqual(data["html"].output, "<p>Hello World!</p>")

        storage["@@plone.app.blocks.richtext/demo"] = {
            "html": RichTextValue(
                "<p>Foo bar!</p>",
                mimeType="text/html",
                outputMimeType="text/x-html-safe",
                encoding="utf-8",
            )
        }

        output = str(storage.storage).replace("\n", "")
        self.assertIn('"html-content-type": "text/html"', output)
        self.assertIn('"html-output-content-type": "text/x-html-safe"', output)
        self.assertIn("<p>Foo bar!</p>", output)



from Acquisition import aq_parent
from plone.app.blocks import utils
from plone.app.blocks.layoutbehavior import ILayoutBehaviorAdaptable
from plone.app.linkintegrity.interfaces import IRetriever
from plone.app.linkintegrity.retriever import DXGeneral
from plone.base.utils import safe_bytes
from plone.tiles.interfaces import ITile
from repoze.xmliter.utils import getHTMLSerializer
from zope.component import adapter
from zope.interface import implementer

import re


@implementer(IRetriever)
@adapter(ILayoutBehaviorAdaptable)
class BlocksDXGeneral(DXGeneral):
    """General retriever for DX that extracts URLs from (rich) text fields."""

    def __init__(self, context):
        self.context = context

    def retrieveLinks(self):
        """Finds all links from the object and return them."""
        links = super(BlocksDXGeneral, self).retrieveLinks()
        links |= self.retrieveLinksFromTiles()
        return links

    def retrieveLinksFromTiles(self):
        links = set()

        if self.context.customContentLayout is None:
            return links

        if aq_parent(self.context) is None or not hasattr(self.context, "REQUEST"):
            # context has not been added to a container yet.
            # This happens when pasting an item.
            # This easily leads to errors traversing to tiles.
            # Return the empty set for now.
            # This code will be triggered again shortly after by another event.
            return links

        iterable = [
            re.sub(b"&#13;", b"\n", re.sub(b"&#13;\n", b"\n", safe_bytes(item)))
            for item in self.context.customContentLayout
            if item
        ]
        result = getHTMLSerializer(
            # safe_bytes (called just before) assumes 'utf-8' encoding
            # thus, we use it here as well
            # if other encoding is needed, those are two places to change
            iterable,
            pretty_print=False,
            encoding="utf-8",
        )

        for tile_node in utils.bodyTileXPath(result.tree):
            tile_url = tile_node.attrib[utils.tileAttrib]
            # assume request query parameters are always useless
            # thus ignore any chars from '?'
            tile_name = tile_url.split("?")[0]
            # eat first two characters of url ('./')
            tile_name = tile_name[2:]
            tile = self.context.restrictedTraverse(tile_name)
            links |= IRetriever(tile).retrieveLinks()

        return links


@implementer(IRetriever)
@adapter(ITile)
class TileGeneral(object):
    def __init__(self, context):
        self.context = context

    def retrieveLinks(self):
        return set()



from plone.app.blocks.interfaces import IBlocksTransformEnabled
from plone.app.blocks.layoutbehavior import ILayoutAware
from plone.app.layout.globals.interfaces import IViewView
from plone.dexterity.browser.view import DefaultView
from plone.outputfilters import apply_filters
from plone.outputfilters.interfaces import IFilter
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.component import getAdapters
from zope.interface import alsoProvides
from zope.interface import implementer

import os


ERROR_LAYOUT = """
<!DOCTYPE html>
<html lang="en" data-layout="./@@page-site-layout">
<body>
    <div data-panel="content">
        Could not find layout for content
    </div>
</body>
</html>"""


@implementer(IViewView)
class SiteLayoutView(BrowserView):
    """Default site layout view called from the site layout resolving view"""

    index = ViewPageTemplateFile(os.path.join("templates", "main_template.pt"))

    def __init__(self, context, request, name="layout"):
        super().__init__(context, request)
        self.__name__ = name

    def __call__(self):
        return self.index()


@implementer(IBlocksTransformEnabled)
class ContentLayoutView(DefaultView):
    """Default view for a layout aware page"""

    def __call__(self):
        """Render the contents of the "content" field coming from
        the ILayoutAware behavior.

        This result is supposed to be transformed by plone.app.blocks.
        """
        lookup = ILayoutAware(self.context, None)
        layout = lookup.content_layout() if lookup else None
        if not layout:
            layout = ERROR_LAYOUT

        # Here we skip legacy portal_transforms and call plone.outputfilters
        # directly by purpose
        filters = [f for _, f in getAdapters((self.context, self.request), IFilter)]
        result = apply_filters(filters, layout)
        if not self.request.response.getHeader("Content-Type"):
            # Needed since Plone 5.2.10.1/6.0.0.1 with Zope security fix.
            self.request.response.setHeader("Content-Type", "text/html")
        return result


@implementer(IBlocksTransformEnabled)
class ContentLayoutPreview(ContentLayoutView):
    """Default view for a layout aware page"""

    def __call__(self):
        from plone.app.drafts.dexterity import IDisplayFormDrafting
        from plone.app.drafts.interfaces import ICurrentDraftManagement
        from plone.app.drafts.proxy import DraftProxy
        from plone.app.drafts.utils import getCurrentDraft

        draft = getCurrentDraft(self.request)
        if draft is not None:
            ICurrentDraftManagement(self.request).mark()
            alsoProvides(self.request, IDisplayFormDrafting)
            self.context = DraftProxy(draft, self.context)

        return super().__call__()


@implementer(IBlocksTransformEnabled)
class TileLayoutView(DefaultView):
    """Introspection view for displaying configured tiles saved in 'content'"""

    def __call__(self):
        """Render the contents of the "content" field coming from
        the ILayoutAware behavior.

        This result is supposed to be transformed by plone.app.blocks.
        """
        lookup = ILayoutAware(self.context, None)
        layout = lookup.tile_layout() if lookup else None
        if not layout:
            layout = ERROR_LAYOUT

        # Here we skip legacy portal_transforms and call plone.outputfilters
        # directly by purpose
        filters = [f for _, f in getAdapters((self.context, self.request), IFilter)]
        result = apply_filters(filters, layout)
        if not self.request.response.getHeader("Content-Type"):
            # Needed since Plone 5.2.10.1/6.0.0.1 with Zope security fix.
            self.request.response.setHeader("Content-Type", "text/html")
        return result



from lxml.html import fromstring
from plone.app.blocks.layoutbehavior import ILayoutAware
from plone.app.blocks.layoutbehavior import ILayoutBehaviorAdaptable
from plone.app.contenttypes import indexers
from plone.base.utils import safe_text
from plone.indexer.decorator import indexer
from plone.tiles.data import ANNOTATIONS_KEY_PREFIX
from zope.annotation.interfaces import IAnnotations
from zope.component import adapter
from zope.interface import implementer


HAS_DEXTERITYTEXTINDEXER = False

try:
    from collective.dexteritytextindexer.interfaces import (  # noqa
        IDynamicTextIndexExtender,
    )

    HAS_DEXTERITYTEXTINDEXER = True
except ImportError:
    pass

try:
    from plone.app.dexterity.textindexer.interfaces import (  # noqa
        IDynamicTextIndexExtender,
    )

    HAS_DEXTERITYTEXTINDEXER = True
except ImportError:
    pass

concat = indexers._unicode_save_string_concat


@indexer(ILayoutBehaviorAdaptable)
def LayoutSearchableText(obj):
    text = [obj.id]
    try:
        text.append(obj.text.output)
    except AttributeError:
        pass
    try:
        text.append(safe_text(obj.title))
    except AttributeError:
        pass
    try:
        text.append(safe_text(obj.description))
    except AttributeError:
        pass

    behavior_data = ILayoutAware(obj)
    # get data from tile data
    annotations = IAnnotations(obj)
    for key in annotations.keys():
        if key.startswith(ANNOTATIONS_KEY_PREFIX):
            data = annotations[key]
            for field_name in ("title", "label", "content"):
                val = data.get(field_name)
                if isinstance(val, str):
                    text.append(val)

    try:
        if behavior_data.content:
            dom = fromstring(behavior_data.content)
            text.extend(dom.xpath("//text()"))
    except AttributeError:
        pass

    try:
        if behavior_data.customLayout:
            dom = fromstring(behavior_data.customLayout)
            text.extend(dom.xpath("//text()"))
    except AttributeError:
        pass

    return concat(*set(text))


if HAS_DEXTERITYTEXTINDEXER:

    @implementer(IDynamicTextIndexExtender)
    @adapter(ILayoutBehaviorAdaptable)
    class LayoutSearchableTextIndexExtender:
        def __init__(self, context):
            self.context = context

        def __call__(self):
            return LayoutSearchableText(self.context)()



from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.registry.interfaces import IRegistry
from plone.testing import Layer
from zope.component import getUtility
from zope.configuration import xmlconfig


class BlocksLayer(PloneSandboxLayer):
    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        # load ZCML
        import plone.app.blocks
        import plone.app.contenttypes
        import plone.app.tiles

        self.loadZCML(package=plone.app.tiles, name="demo.zcml")
        self.loadZCML(package=plone.app.blocks)
        self.loadZCML(package=plone.app.contenttypes)

        # Register directory for testing
        xmlconfig.string(
            """\
<configure
    xmlns="http://namespaces.zope.org/zope"
    xmlns:plone="http://namespaces.plone.org/plone"
    i18n_domain="plone"
    package="plone.app.blocks.tests">

    <plone:static
        type="sitelayout"
        name="testlayout1"
        directory="resources/sitelayout/testlayout1"
        />

    <plone:static
        type="sitelayout"
        name="testlayout2"
        directory="resources/sitelayout/testlayout2"
        />

    <plone:static
        type="contentlayout"
        name="testlayout1"
        directory="resources/contentlayout/testlayout1"
        />

    <plone:static
        type="contentlayout"
        name="testlayout2"
        directory="resources/contentlayout/testlayout2"
        />

</configure>
""",
            context=configurationContext,
        )
        if "virtual_hosting" not in app.objectIds():
            # If ZopeLite was imported, we have no default virtual
            # host monster
            from Products.SiteAccess.VirtualHostMonster import (
                manage_addVirtualHostMonster,
            )

            manage_addVirtualHostMonster(app, "virtual_hosting")

    def setUpPloneSite(self, portal):
        # ensure plone.app.theming disabled
        registry = getUtility(IRegistry)
        key = "plone.app.theming.interfaces.IThemeSettings.enabled"
        if key in registry:
            registry[key] = False
        # install plone.app.contenttypes on Plone 5
        self.applyProfile(portal, "plone.app.contenttypes:default")
        # install into the Plone site
        self.applyProfile(portal, "plone.app.blocks:default")


class PrettyPrintLayer(Layer):
    def setUp(self):
        from plone.app.blocks.transform import ParseXML

        ParseXML.pretty_print = True

    def tearDown(self):
        from plone.app.blocks.transform import ParseXML

        ParseXML.pretty_print = False


BLOCKS_FIXTURE = BlocksLayer()
PRETTY_PRINT_FIXTURE = PrettyPrintLayer()
BLOCKS_INTEGRATION_TESTING = IntegrationTesting(
    bases=(BLOCKS_FIXTURE,), name="Blocks:Integration"
)
BLOCKS_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(BLOCKS_FIXTURE,), name="Blocks:Functional"
)
BLOCKS_FUNCTIONAL_TESTING_PRETTY_PRINT = FunctionalTesting(
    bases=(
        PRETTY_PRINT_FIXTURE,
        BLOCKS_FIXTURE,
    ),
    name="Blocks:FunctionalPrettyPrinting",
)



from plone.app.blocks.interfaces import DEFAULT_SITE_LAYOUT_REGISTRY_KEY
from plone.app.blocks.layoutbehavior import ILayoutAware
from plone.app.blocks.layoutbehavior import LayoutAwareBehavior
from plone.app.blocks.testing import BLOCKS_FUNCTIONAL_TESTING
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.registry.interfaces import IRegistry
from zExceptions import NotFound
from zope.component import getGlobalSiteManager
from zope.component import getMultiAdapter
from zope.component import getUtility

import transaction
import unittest


class TestPageSiteLayout(unittest.TestCase):
    layer = BLOCKS_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self.registry = getUtility(IRegistry)

        setRoles(self.portal, TEST_USER_ID, ("Manager",))
        self.portal.invokeFactory("Folder", "f1", title="Folder 1")
        self.portal["f1"].invokeFactory("Document", "d1", title="Document 1")
        setRoles(self.portal, TEST_USER_ID, ("Member",))

        # setup default behavior
        self.registry[
            DEFAULT_SITE_LAYOUT_REGISTRY_KEY
        ] = b"/++sitelayout++testlayout1/site.html"
        iface = self.portal["f1"]["d1"].__class__
        sm = getGlobalSiteManager()
        sm.registerAdapter(LayoutAwareBehavior, [iface])
        registrations = sm.getAdapters((self.portal["f1"]["d1"],), ILayoutAware)
        self.assertEqual(len(list(registrations)), 1)
        self.behavior = ILayoutAware(self.portal["f1"]["d1"])
        registrations = sm.getAdapters((self.portal["f1"]["d1"],), ILayoutAware)
        self.assertEqual(len(list(registrations)), 1)

    def test_page_site_layout_no_registry_key(self):
        self.registry[DEFAULT_SITE_LAYOUT_REGISTRY_KEY] = None
        view = getMultiAdapter(
            (
                self.portal["f1"]["d1"],
                self.request,
            ),
            name="page-site-layout",
        )
        self.assertRaises(NotFound, view.index)

        from plone.app.blocks.layoutviews import SiteLayoutView

        default_view = SiteLayoutView(self.portal["f1"]["d1"], self.request)
        self.assertEqual(view().split(), default_view().split())

    def test_page_site_layout_default(self):
        self.registry[
            DEFAULT_SITE_LAYOUT_REGISTRY_KEY
        ] = b"/++sitelayout++testlayout1/site.html"
        view = getMultiAdapter(
            (
                self.portal["f1"]["d1"],
                self.request,
            ),
            name="page-site-layout",
        )
        rendered = view()
        self.assertTrue("Layout title" in rendered)

    def test_page_site_layout_page_override(self):
        self.behavior.customContentLayout = "<html><body>N/A</body></html>"
        self.behavior.pageSiteLayout = "/++sitelayout++testlayout2/mylayout.html"

        view = getMultiAdapter(
            (
                self.portal["f1"]["d1"],
                self.request,
            ),
            name="page-site-layout",
        )
        rendered = view()
        self.assertFalse("Layout title" in rendered)
        self.assertTrue("My Layout 1 Title" in rendered)

    def test_page_site_layout_section_override(self):
        self.behavior.sectionSiteLayout = "/++sitelayout++testlayout2/mylayout.html"

        view = getMultiAdapter(
            (
                self.portal["f1"]["d1"],
                self.request,
            ),
            name="page-site-layout",
        )
        rendered = view()

        self.assertFalse("Layout title" in rendered)
        self.assertTrue("My Layout 1 Title" in rendered)

    def test_page_site_layout_cache(self):
        self.behavior.pageSiteLayout = "/++sitelayout++testlayout2/mylayout.html"

        view = getMultiAdapter(
            (
                self.portal["f1"]["d1"],
                self.request,
            ),
            name="page-site-layout",
        )
        rendered = view()

        self.assertFalse("Layout title" in rendered)
        self.assertTrue("My Layout 1 Title" in rendered)

        # Change the section value
        self.behavior.pageSiteLayout = "/++sitelayout++testlayout1/site.html"

        view = getMultiAdapter(
            (
                self.portal["f1"]["d1"],
                self.request,
            ),
            name="page-site-layout",
        )
        rendered = view()

        # Cache means our change is ignored
        self.assertFalse("Layout title" in rendered)
        self.assertTrue("My Layout 1 Title" in rendered)

    def test_page_site_layout_cache_invalidate_mtime(self):
        self.behavior.customContentLayout = "<html><body>N/A</body></html>"
        self.behavior.sectionSiteLayout = "/++sitelayout++testlayout2/mylayout.html"

        view = getMultiAdapter(
            (
                self.portal["f1"]["d1"],
                self.request,
            ),
            name="page-site-layout",
        )
        rendered = view()

        self.assertFalse("Layout title" in rendered)
        self.assertTrue("My Layout 1 Title" in rendered)

        # Trigger invalidation by modifying the context
        self.portal["f1"]["d1"].title = "New title"
        transaction.commit()

        # Change the section value
        self.behavior.sectionSiteLayout = "/++sitelayout++testlayout1/site.html"

        view = getMultiAdapter(
            (
                self.portal["f1"]["d1"],
                self.request,
            ),
            name="page-site-layout",
        )
        rendered = view()

        # We now get the new layout
        self.assertTrue("Layout title" in rendered)
        self.assertFalse("My Layout 1 Title" in rendered)

    def test_page_site_layout_cache_invalidate_catalog_counter(self):
        self.behavior.sectionSiteLayout = "/++sitelayout++testlayout2/mylayout.html"

        view = getMultiAdapter(
            (
                self.portal["f1"]["d1"],
                self.request,
            ),
            name="page-site-layout",
        )
        rendered = view()

        self.assertFalse("Layout title" in rendered)
        self.assertTrue("My Layout 1 Title" in rendered)

        # Trigger invalidation by incrementing the catalog counter
        self.portal["portal_catalog"]._increment_counter()

        # Change the section value
        self.behavior.sectionSiteLayout = "/++sitelayout++testlayout1/site.html"

        view = getMultiAdapter(
            (
                self.portal["f1"]["d1"],
                self.request,
            ),
            name="page-site-layout",
        )
        rendered = view()

        # We now get the new layout
        self.assertTrue("Layout title" in rendered)
        self.assertFalse("My Layout 1 Title" in rendered)

    def test_page_site_layout_cache_invalidate_registry_key(self):
        self.behavior.sectionSiteLayout = "/++sitelayout++testlayout2/mylayout.html"

        view = getMultiAdapter(
            (
                self.portal["f1"]["d1"],
                self.request,
            ),
            name="page-site-layout",
        )
        rendered = view()

        self.assertFalse("Layout title" in rendered)
        self.assertTrue("My Layout 1 Title" in rendered)

        # Trigger invalidation by modifying the global registry key
        self.registry[
            DEFAULT_SITE_LAYOUT_REGISTRY_KEY
        ] = b"/++sitelayout++testlayout2/mylayout.html"

        # Change the section value
        self.behavior.sectionSiteLayout = "/++sitelayout++testlayout1/site.html"

        view = getMultiAdapter(
            (
                self.portal["f1"]["d1"],
                self.request,
            ),
            name="page-site-layout",
        )
        rendered = view()

        # We now get the new layout
        self.assertTrue("Layout title" in rendered)
        self.assertFalse("My Layout 1 Title" in rendered)


class TestPageSiteLayoutAcquisition(unittest.TestCase):
    layer = BLOCKS_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self.registry = getUtility(IRegistry)

        setRoles(self.portal, TEST_USER_ID, ("Manager",))
        self.portal.invokeFactory("Folder", "f1", title="Folder 1")
        self.portal["f1"].invokeFactory("Document", "d1", title="Document 1")
        setRoles(self.portal, TEST_USER_ID, ("Member",))

        sm = getGlobalSiteManager()

        # setup default behaviors
        self.registry[
            DEFAULT_SITE_LAYOUT_REGISTRY_KEY
        ] = b"/++sitelayout++testlayout1/site.html"

        iface = self.portal["f1"].__class__
        sm.registerAdapter(LayoutAwareBehavior, [iface])

        iface = self.portal["f1"]["d1"].__class__
        sm.registerAdapter(LayoutAwareBehavior, [iface])

        registrations = sm.getAdapters((self.portal["f1"],), ILayoutAware)
        self.assertEqual(len(list(registrations)), 1)

        registrations = sm.getAdapters((self.portal["f1"]["d1"],), ILayoutAware)
        self.assertEqual(len(list(registrations)), 1)

    def test_page_site_layout_is_not_acquired(self):
        self.registry[
            DEFAULT_SITE_LAYOUT_REGISTRY_KEY
        ] = b"/++sitelayout++testlayout1/site.html"

        a1 = ILayoutAware(self.portal["f1"])
        a2 = ILayoutAware(self.portal["f1"]["d1"])

        self.assertEqual(a1.site_layout(), a2.site_layout())

        a1.pageSiteLayout = "/++sitelayout++testlayout2/mylayout.html"

        self.assertNotEqual(a1.site_layout(), a2.site_layout())



from plone.registry import field
from plone.registry import Record
from plone.registry.interfaces import IRegistry
from zope.component import getUtility


def step_setup_various(context):
    if context.readDataFile("plone.app.blocks_default.txt") is None:
        return
    portal = context.getSite()
    initialize_default_layout_registry_values(portal)


def initialize_default_layout_registry_values(portal):
    registry = getUtility(IRegistry)
    records = (
        (
            "plone.defaultSiteLayout",
            "Default site layout",
            "The default site layout for the site",
            None,
        ),
        (
            "plone.defaultAjaxLayout",
            "Default ajax layout",
            "The default ajax layout for the site",
            None,
        ),
    )
    for key, title, description, value in records:
        if key not in registry.records:
            registry.records[key] = Record(
                field.BytesLine(title=title, description=description, required=False),
                value,
            )



from lxml import html
from lxml.etree import XSLTApplyError
from plone.app.blocks import events
from plone.app.blocks import PloneMessageFactory
from plone.app.blocks import utils
from plone.app.blocks.interfaces import IBlocksSettings
from plone.app.blocks.utils import resolve_transform
from plone.registry.interfaces import IRegistry
from plone.tiles.interfaces import ESI_HEADER
from plone.tiles.interfaces import ESI_HEADER_KEY
from urllib import parse
from zExceptions import NotFound
from zope.component import queryUtility
from zope.event import notify
from zope.i18n import translate

import logging


logger = logging.getLogger(__name__)


def errorTile(request):
    msg = PloneMessageFactory("There was an error while rendering this tile")
    translated = translate(msg, context=request)
    return html.fromstring(translated).getroottree()


def renderTiles(request, tree):
    """Find all tiles in the given response, contained in the lxml element
    tree `tree`, and insert them into the output.

    Assumes panel merging has already happened.
    """
    # Optionally enable ESI rendering in tiles that support this
    if not request.getHeader(ESI_HEADER):
        registry = queryUtility(IRegistry)
        if registry is not None:
            if registry.forInterface(IBlocksSettings, check=False).esi:
                request.environ[ESI_HEADER_KEY] = "true"

    root = tree.getroot()
    headNode = root.find("head")
    baseURL = request.getURL()
    if request.getVirtualRoot():
        # plone.subrequest deals with VHM requests
        baseURL = ""
    for tileNode in utils.headTileXPath(tree):
        tileHref = tileNode.attrib[utils.tileAttrib]
        if not tileHref.startswith("/"):
            tileHref = parse.urljoin(baseURL, tileHref)

        notify(events.BeforeTileRenderEvent(tileHref, tileNode))

        tileTree = None
        try:
            tileTree = utils.resolve(tileHref)
        except RuntimeError:
            tileTree = None
        except NotFound:
            logger.warn("NotFound while trying to render tile: %s", tileHref)

        if tileTree is None:
            utils.remove_element(tileNode)
            notify(events.AfterTileRenderEvent(tileHref, tileNode))
            continue

        if tileTree is not None:
            tileRoot = tileTree.getroot()
            tileHead = tileRoot.find("head")
            if tileHead is None and tileRoot.find("body") is None:
                tileHead = tileRoot

            utils.replace_with_children(tileNode, tileHead)

        notify(events.AfterTileRenderEvent(tileHref, tileNode))

    for tileNode in utils.bodyTileXPath(tree):
        tileHref = tileNode.attrib[utils.tileAttrib]
        tileRulesHref = tileNode.attrib.get(utils.tileRulesAttrib)

        if not tileHref.startswith("/"):
            tileHref = parse.urljoin(baseURL, tileHref)

        notify(events.BeforeTileRenderEvent(tileHref, tileNode))

        tileTree = None
        try:
            tileTree = utils.resolve(tileHref)
        except RuntimeError:
            tileTree = errorTile(request)
        except NotFound:
            logger.warning("NotFound while trying to render tile: %s", tileHref)

        if tileTree is None:
            utils.remove_element(tileNode)
            notify(events.AfterTileRenderEvent(tileHref, tileNode))
            continue

        tileTransform = None
        if tileRulesHref:
            if not tileRulesHref.startswith("/"):
                tileRulesHref = parse.urljoin(baseURL, tileRulesHref)
            try:
                tileTransform = resolve_transform(tileRulesHref, tileNode)
            except NotFound:
                tileTransform = None
            del tileNode.attrib[utils.tileRulesAttrib]

        if tileTree is not None:
            tileRoot = tileTree.getroot()
            tileHead = tileRoot.find("head")
            tileBody = tileRoot.find("body")

            if tileHead is None and tileBody is None:
                tileBody = tileRoot

            if tileTransform is not None:
                try:
                    result = tileTransform(tileBody).getroot()
                    del tileBody[:]
                    tileBody.append(result)
                except XSLTApplyError:
                    logger.exception(
                        "Failed to transform tile %s for %s", tileHref, baseURL
                    )

            if tileHead is not None:
                for tileHeadChild in tileHead:
                    headNode.append(tileHeadChild)

            utils.replace_with_children(tileNode, tileBody)

        notify(events.AfterTileRenderEvent(tileHref, tileNode))

    if utils.headTileXPath(tree) or utils.bodyTileXPath(tree):
        # Recursively render tiles, as long as they are referenced
        tree = renderTiles(request, tree)

    return tree



from io import StringIO
from lxml import etree
from lxml import html
from OFS.Image import File
from plone.app.blocks.interfaces import DEFAULT_SITE_LAYOUT_REGISTRY_KEY
from plone.app.blocks.panel import merge
from plone.app.blocks.testing import BLOCKS_FUNCTIONAL_TESTING
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.memoize.volatile import ATTR
from plone.registry.interfaces import IRegistry
from Products.BTreeFolder2.BTreeFolder2 import BTreeFolder2
from Products.CMFCore.utils import getToolByName
from zExceptions import NotFound
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.component import getSiteManager
from zope.component import getUtility
from zope.component import provideAdapter
from zope.interface import implementer
from zope.interface import Interface
from zope.publisher.browser import BrowserPage
from zope.publisher.interfaces.browser import IBrowserPage

import transaction
import unittest


class TestSiteLayout(unittest.TestCase):
    layer = BLOCKS_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self.registry = getUtility(IRegistry)
        setRoles(self.portal, TEST_USER_ID, ("Manager",))
        self.portal.invokeFactory("Folder", "f1", title="Folder 1")
        self.portal["f1"].invokeFactory("Document", "d1", title="Document 1")
        setRoles(self.portal, TEST_USER_ID, ("Member",))

    def test_default_site_layout_no_registry_key(self):
        # Clear cache if there
        if hasattr(self.portal, ATTR):
            delattr(self.portal, ATTR)

        self.registry[DEFAULT_SITE_LAYOUT_REGISTRY_KEY] = None

        view = getMultiAdapter(
            (
                self.portal,
                self.request,
            ),
            name="default-site-layout",
        )
        self.assertRaises(NotFound, view.index)

        from plone.app.blocks.layoutviews import SiteLayoutView

        default_view = SiteLayoutView(self.portal, self.request)
        self.assertEqual(view().split(), default_view().split())

    def test_default_site_layout(self):
        # Clear cache if there
        if hasattr(self.portal, ATTR):
            delattr(self.portal, ATTR)

        self.registry[
            DEFAULT_SITE_LAYOUT_REGISTRY_KEY
        ] = b"/++sitelayout++testlayout1/site.html"

        view = getMultiAdapter(
            (
                self.portal,
                self.request,
            ),
            name="default-site-layout",
        )
        rendered = view()

        self.assertTrue("Layout title" in rendered)

    def test_no_default_site_layout(self):
        # Clear cache if there
        if hasattr(self.portal, ATTR):
            delattr(self.portal, ATTR)

        view = getMultiAdapter(
            (
                self.portal,
                self.request,
            ),
            name="default-site-layout",
        )
        rendered = view()

        # Should render main_template with template-layout in body class
        rendered_tree = etree.parse(StringIO(rendered), etree.HTMLParser())
        xpath_body = etree.XPath("/html/body")
        body_tag = xpath_body(rendered_tree)[0]
        self.assertIn("template-layout", body_tag.attrib["class"])

    def test_default_site_layout_section_override(self):
        self.registry[
            DEFAULT_SITE_LAYOUT_REGISTRY_KEY
        ] = b"/++sitelayout++testlayout1/site.html"

        from plone.app.blocks.layoutbehavior import ILayoutAware

        @implementer(ILayoutAware)
        @adapter(self.portal["f1"].__class__)
        class FolderLayoutAware:
            def __init__(self, context):
                self.context = context

            customContentLayout = "<html><body>N/A</body></html>"
            sectionSiteLayout = "/++sitelayout++testlayout2/mylayout.html"
            pageSiteLayout = None

        # Register a local adapter for easy rollback
        sm = getSiteManager()
        sm.registerAdapter(FolderLayoutAware)

        view = getMultiAdapter(
            (
                self.portal["f1"]["d1"],
                self.request,
            ),
            name="default-site-layout",
        )
        rendered = view()

        self.assertFalse("Layout title" in rendered)
        self.assertTrue("My Layout 1 Title" in rendered)

    def test_default_site_layout_section_no_override(self):
        self.registry[
            DEFAULT_SITE_LAYOUT_REGISTRY_KEY
        ] = b"/++sitelayout++testlayout1/site.html"

        view = getMultiAdapter(
            (
                self.portal["f1"]["d1"],
                self.request,
            ),
            name="default-site-layout",
        )
        rendered = view()

        self.assertFalse("My Layout 1 Title" in rendered)
        self.assertTrue("Layout title" in rendered)

    def test_default_site_layout_cache(self):
        # Clear cache if there
        if hasattr(self.portal, ATTR):
            delattr(self.portal, ATTR)

        resources = getToolByName(self.portal, "portal_resources")
        resources._setOb("sitelayout", BTreeFolder2("sitelayout"))
        resources["sitelayout"]._setOb("testlayout3", BTreeFolder2("testlayout3"))
        resources["sitelayout"]["testlayout3"]._setOb(
            "site.html",
            File(
                "site.html",
                "site.html",
                b"<html><head><title>ZODB test</title></head></html>",
            ),
        )

        self.registry[
            DEFAULT_SITE_LAYOUT_REGISTRY_KEY
        ] = b"/++sitelayout++testlayout3/site.html"

        view = getMultiAdapter(
            (
                self.portal,
                self.request,
            ),
            name="default-site-layout",
        )
        rendered = view()

        self.assertTrue("ZODB test" in rendered)

        resources["sitelayout"]["testlayout3"]._delOb("site.html")
        resources["sitelayout"]["testlayout3"]._setOb(
            "site.html",
            File(
                "site.html",
                "site.html",
                b"<html><head><title>Cache test</title></head></html>",
            ),
        )

        view = getMultiAdapter(
            (
                self.portal,
                self.request,
            ),
            name="default-site-layout",
        )
        rendered = view()

        self.assertFalse("Cache test" in rendered)  # hidden by cache
        self.assertTrue("ZODB test" in rendered)

        self.assertEqual("/++sitelayout++testlayout3/site.html", view.layout)

        # Test cache is set
        self.assertTrue(hasattr(self.portal, ATTR))

        # Update cache
        for key in getattr(self.portal, ATTR):
            getattr(self.portal, ATTR)[key] = None

        self.assertIsNone(view.layout)  # because of cache

    def test_default_site_layout_invalidate_mtime(self):
        # Clear cache if there
        if hasattr(self.portal, ATTR):
            delattr(self.portal, ATTR)

        resources = getToolByName(self.portal, "portal_resources")
        resources._setOb("sitelayout", BTreeFolder2("sitelayout"))
        resources["sitelayout"]._setOb("testlayout3", BTreeFolder2("testlayout3"))
        resources["sitelayout"]["testlayout3"]._setOb(
            "site.html",
            File(
                "site.html",
                "site.html",
                b"<html><head><title>ZODB test</title></head></html>",
            ),
        )

        self.registry[
            DEFAULT_SITE_LAYOUT_REGISTRY_KEY
        ] = b"/++sitelayout++testlayout3/site.html"

        view = getMultiAdapter(
            (
                self.portal,
                self.request,
            ),
            name="default-site-layout",
        )
        rendered = view()

        self.assertTrue("ZODB test" in rendered)

        # Trigger invalidation by modifying the context and committing
        self.portal.title = "New title"
        transaction.commit()

        # Modify the site layout
        resources["sitelayout"]["testlayout3"]._delOb("site.html")
        resources["sitelayout"]["testlayout3"]._setOb(
            "site.html",
            File(
                "site.html",
                "site.html",
                b"<html><head><title>Cache test</title></head></html>",
            ),
        )

        view = getMultiAdapter(
            (
                self.portal,
                self.request,
            ),
            name="default-site-layout",
        )
        rendered = view()

        self.assertTrue("Cache test" in rendered)
        self.assertFalse("ZODB test" in rendered)

    def test_default_site_layout_invalidate_registry_key(self):
        # Clear cache if there
        if hasattr(self.portal, ATTR):
            delattr(self.portal, ATTR)

        self.registry[
            DEFAULT_SITE_LAYOUT_REGISTRY_KEY
        ] = b"/++sitelayout++testlayout1/site.html"

        view = getMultiAdapter(
            (
                self.portal,
                self.request,
            ),
            name="default-site-layout",
        )
        rendered = view()

        self.assertTrue("Layout title" in rendered)

        # Trigger invalidation by modifying the global site layout selection
        self.registry[
            DEFAULT_SITE_LAYOUT_REGISTRY_KEY
        ] = b"/++sitelayout++testlayout2/mylayout.html"

        view = getMultiAdapter(
            (
                self.portal,
                self.request,
            ),
            name="default-site-layout",
        )
        rendered = view()

        self.assertTrue("My Layout 1 Title" in rendered)
        self.assertFalse("Layout title" in rendered)

    def test_panel_modes(self):
        """Test data-panel-mode ``append`` and ``replace``."""

        site_layout = """
          <!DOCTYPE html>
          <html>
            <head>
              <title>okayish site layout</title>
            </head>
            <body>
              <div class="panel-1" data-panel="content-1"/>
              <div class="panel-2" data-panel="content-2" data-panel-mode="append"/>
              <div class="panel-3" data-panel="content-3" data-panel-mode="replace"/>
            </body>
          </html>
        """

        content_layout = """
          <html data-layout="./@@testsitelayout">
            <body>
              <div data-panel="content-1">
                <article class="content-1">content 1</article>
              </div>
              <div data-panel="content-2">
                <article class="content-2">content 2</article>
              </div>
              <div data-panel="content-3">
                <article class="content-3">content 3</article>
              </div>
            </body>
          </html>
        """

        class SiteLayout(BrowserPage):
            def __call__(self):
                return site_layout

        provideAdapter(
            SiteLayout,
            adapts=(Interface, Interface),
            provides=IBrowserPage,
            name="testsitelayout",
        )

        parser = html.HTMLParser(encoding="utf-8")
        page_tree = html.fromstring(content_layout, parser=parser).getroottree()

        merged = merge(self.request, page_tree)

        # Default case: append matched content panel into layout panel.
        r1 = merged.xpath(
            "//div[contains(@class, 'panel-1')]/article[contains(@class, 'content-1')]"
        )
        self.assertEqual(len(r1), 1)

        # Mode "append": append matched content panel into layout panel.
        r2 = merged.xpath(
            "//div[contains(@class, 'panel-2')]/article[contains(@class, 'content-2')]"
        )
        self.assertEqual(len(r2), 1)

        # Mode "replace": replace layout panel with matched content panel.
        r3 = merged.xpath("//div[contains(@class, 'panel-3')]")
        self.assertEqual(len(r3), 0)
        r4 = merged.xpath("//article[contains(@class, 'content-3')]")
        self.assertEqual(len(r4), 1)
